const coral = @import("coral");

const ext = @cImport({
	@cInclude("spirv-cross/spirv_cross_c.h");
});

const sokol = @import("sokol");

const std = @import("std");

pub const Error = std.mem.Allocator.Error || error {
	UnsupportedTarget,
	InvalidSPIRV,
	UnsupportedSPIRV,
};

pub const Unit = struct {
	context: ext.spvc_context,
	shader_desc: sokol.gfx.ShaderDesc,
	attrs_used: u32 = 0,

	pub fn compile(self: *Unit, spirv: []const u32, stage: Stage) Error!void {
		const execution_model, const stage_desc = switch (stage) {
			.vertex => .{ext.SpvExecutionModelVertex, &self.shader_desc.vs},
			.fragment => .{ext.SpvExecutionModelFragment, &self.shader_desc.fs},
		};

		const Backend = struct {
			target: ext.spvc_backend,
			option_values: []const struct {ext.spvc_compiler_option, c_uint},
		};

		const backend: Backend = switch (sokol.gfx.queryBackend()) {
			.GLCORE => .{
				.target = ext.SPVC_BACKEND_GLSL,

				.option_values = &.{
					.{ext.SPVC_COMPILER_OPTION_GLSL_VERSION, 430},
					.{ext.SPVC_COMPILER_OPTION_GLSL_ES, @intFromBool(false)},
					.{ext.SPVC_COMPILER_OPTION_GLSL_VULKAN_SEMANTICS, @intFromBool(false)},
					.{ext.SPVC_COMPILER_OPTION_GLSL_EMIT_UNIFORM_BUFFER_AS_PLAIN_UNIFORMS, @intFromBool(true)},
				},
			},

			else => @panic("Unimplemented"),
		};

		const compiler = parse_and_configure: {
			var parsed_ir: ext.spvc_parsed_ir = null;

			try to_error(ext.spvc_context_parse_spirv(self.context, spirv.ptr, spirv.len, &parsed_ir));

			var compiler: ext.spvc_compiler = null;

			try to_error(ext.spvc_context_create_compiler(self.context, backend.target, parsed_ir, ext.SPVC_CAPTURE_MODE_TAKE_OWNERSHIP, &compiler));
			try to_error(ext.spvc_compiler_build_combined_image_samplers(compiler));

			var combined_image_samplers: []const ext.spvc_combined_image_sampler = &.{};

			try to_error(ext.spvc_compiler_get_combined_image_samplers(compiler, @ptrCast(&combined_image_samplers.ptr), &combined_image_samplers.len));

			var binding: u32 = 0;

			for (combined_image_samplers) |combined_image_sampler| {
				var name_buffer = [_:0]u8{0} ** 255;

				const name = coral.utf8.print_formatted(&name_buffer, "{image_name}_{sampler_name}", .{
					.image_name = coral.io.slice_sentineled(u8, 0, ext.spvc_compiler_get_name(compiler, combined_image_sampler.image_id)),
					.sampler_name = coral.io.slice_sentineled(u8, 0, ext.spvc_compiler_get_name(compiler, combined_image_sampler.sampler_id)),
				}) catch {
					return error.InvalidSPIRV;
				};

				ext.spvc_compiler_set_name(compiler, combined_image_sampler.combined_id, name);
				ext.spvc_compiler_set_decoration(compiler, combined_image_sampler.combined_id, ext.SpvDecorationBinding, binding);

				binding += 1;
			}

			break: parse_and_configure compiler;
		};

		try to_error(ext.spvc_compiler_set_entry_point(compiler, stage_desc.entry, @intCast(execution_model)));

		const resources = create: {
			var resources: ext.spvc_resources = null;

			try to_error(ext.spvc_compiler_create_shader_resources(compiler, &resources));

			break: create resources;
		};

		try reflect_uniform_blocks(compiler, resources, stage_desc);
		try reflect_image_samplers(compiler, resources, stage_desc);

		try to_error(ext.spvc_compiler_install_compiler_options(compiler, create: {
			var options: ext.spvc_compiler_options = null;

			try to_error(ext.spvc_compiler_create_compiler_options(compiler, &options));

			for (backend.option_values) |option_value| {
				const entry, const value = option_value;

				try to_error(ext.spvc_compiler_options_set_uint(options, entry, value));
			}

			break: create options;
		}));

		try to_error(ext.spvc_compiler_compile(compiler, @ptrCast(&stage_desc.source)));
	}

	pub fn deinit(self: *Unit) void {
		ext.spvc_context_destroy(self.context);

		self.* = undefined;
	}

	pub fn init() std.mem.Allocator.Error!Unit {
		var context: ext.spvc_context = null;

		if (ext.spvc_context_create(&context) != ext.SPVC_SUCCESS) {
			return error.OutOfMemory;
		}

		errdefer ext.spvc_context_destroy(context);

		ext.spvc_context_set_error_callback(context, log_context_errors, null);

		return .{
			.context = context,

			.shader_desc = .{
				.vs = .{.entry = "main"},
				.fs = .{.entry = "main"},
			},
		};
	}
};

pub const Stage = enum {
	fragment,
	vertex,
};

fn log_context_errors(userdata: ?*anyopaque, error_message: [*c]const u8) callconv(.C) void {
	std.debug.assert(userdata == null);
	std.log.err("{s}", .{error_message});
}

fn reflect_image_samplers(compiler: ext.spvc_compiler, resources: ext.spvc_resources, stage: *sokol.gfx.ShaderStageDesc) Error!void {
	var reflected_sampled_images: []const ext.spvc_reflected_resource = &.{};

	try to_error(ext.spvc_resources_get_resource_list_for_type(
		resources,
		ext.SPVC_RESOURCE_TYPE_SAMPLED_IMAGE,
		@ptrCast(&reflected_sampled_images.ptr),
		&reflected_sampled_images.len,
	));

	if (reflected_sampled_images.len > stage.image_sampler_pairs.len) {
		return error.UnsupportedSPIRV;
	}

	for (0 .. reflected_sampled_images.len, reflected_sampled_images) |i, reflected_sampled_image| {
		const sampled_image_type = ext.spvc_compiler_get_type_handle(compiler, reflected_sampled_image.type_id);

		if (ext.spvc_type_get_basetype(sampled_image_type) != ext.SPVC_BASETYPE_SAMPLED_IMAGE) {
			return error.InvalidSPIRV;
		}

		stage.images[i] = .{
			.multisampled = ext.spvc_type_get_image_multisampled(sampled_image_type) != 0,

			.image_type = try switch (ext.spvc_type_get_image_dimension(sampled_image_type)) {
				ext.SpvDim2D => sokol.gfx.ImageType._2D,
				else => error.InvalidSPIRV,
			},

			.sample_type = try switch (ext.spvc_type_get_basetype(ext.spvc_compiler_get_type_handle(compiler, ext.spvc_type_get_image_sampled_type(sampled_image_type)))) {
				ext.SPVC_BASETYPE_FP32 => sokol.gfx.ImageSampleType.FLOAT,
				else => error.InvalidSPIRV,
			},

			.used = true,
		};

		stage.samplers[i] = .{
			.sampler_type = .DEFAULT,
			.used = true,
		};

		stage.image_sampler_pairs[i] = .{
			.glsl_name = ext.spvc_compiler_get_name(compiler, reflected_sampled_image.id),
			.image_slot = @intCast(i),
			.sampler_slot = @intCast(i),
			.used = true,
		};
	}
}

fn reflect_uniform_blocks(compiler: ext.spvc_compiler, resources: ext.spvc_resources, stage: *sokol.gfx.ShaderStageDesc) Error!void {
	var reflected_uniform_buffers: []const ext.spvc_reflected_resource = &.{};

	try to_error(ext.spvc_resources_get_resource_list_for_type(
		resources,
		ext.SPVC_RESOURCE_TYPE_UNIFORM_BUFFER,
		@ptrCast(&reflected_uniform_buffers.ptr),
		&reflected_uniform_buffers.len,
	));

	if (reflected_uniform_buffers.len > stage.uniform_blocks.len) {
		return error.UnsupportedSPIRV;
	}

	for (stage.uniform_blocks[0 .. reflected_uniform_buffers.len], reflected_uniform_buffers) |*uniform_block, reflected_uniform_buffer| {
		const uniform_buffer_type = ext.spvc_compiler_get_type_handle(compiler, reflected_uniform_buffer.type_id);

		if (ext.spvc_type_get_basetype(uniform_buffer_type) != ext.SPVC_BASETYPE_STRUCT) {
			return error.InvalidSPIRV;
		}

		const member_count = ext.spvc_type_get_num_member_types(uniform_buffer_type);

		if (member_count > uniform_block.uniforms.len) {
			return error.UnsupportedSPIRV;
		}

		try to_error(ext.spvc_compiler_get_declared_struct_size(compiler, uniform_buffer_type, &uniform_block.size));

		var uniform_blocks_used: u32 = 0;

		while (uniform_blocks_used < member_count) : (uniform_blocks_used += 1) {
			const member_type_id = ext.spvc_type_get_member_type(uniform_buffer_type, uniform_blocks_used);
			const member_type = ext.spvc_compiler_get_type_handle(compiler, member_type_id);

			uniform_block.uniforms[uniform_blocks_used] = .{
				// .name = ext.spvc_compiler_get_member_name(compiler, ext.spvc_type_get_base_type_id(uniform_buffer_type), uniform_blocks_used),

				.array_count = @intCast(try switch (ext.spvc_type_get_num_array_dimensions(member_type)) {
					0 => 0,

					1 => switch (ext.spvc_type_array_dimension_is_literal(member_type, 1) != 0) {
						true => ext.spvc_type_get_array_dimension(member_type, 1),
						false => error.InvalidSPIRV,
					},

					else => error.InvalidSPIRV,
				}),

				.type = try switch (ext.spvc_type_get_basetype(member_type)) {
					ext.SPVC_BASETYPE_INT32 => switch (ext.spvc_type_get_vector_size(member_type)) {
						1 => if (ext.spvc_type_get_columns(member_type) == 1) sokol.gfx.UniformType.INT else error.InvalidSPIRV,
						2 => if (ext.spvc_type_get_columns(member_type) == 1) sokol.gfx.UniformType.INT2 else error.InvalidSPIRV,
						3 => if (ext.spvc_type_get_columns(member_type) == 1) sokol.gfx.UniformType.INT3 else error.InvalidSPIRV,
						4 => if (ext.spvc_type_get_columns(member_type) == 1) sokol.gfx.UniformType.INT4 else error.InvalidSPIRV,
						else => error.InvalidSPIRV,
					},

					ext.SPVC_BASETYPE_FP32 => switch (ext.spvc_type_get_vector_size(member_type)) {
						1 => if (ext.spvc_type_get_columns(member_type) == 1) sokol.gfx.UniformType.FLOAT else error.InvalidSPIRV,
						2 => if (ext.spvc_type_get_columns(member_type) == 1) sokol.gfx.UniformType.FLOAT2 else error.InvalidSPIRV,
						3 => if (ext.spvc_type_get_columns(member_type) == 1) sokol.gfx.UniformType.FLOAT3 else error.InvalidSPIRV,

						4 => switch (ext.spvc_type_get_columns(member_type)) {
							1 => sokol.gfx.UniformType.FLOAT4,
							4 => sokol.gfx.UniformType.MAT4,
							else => error.InvalidSPIRV,
						},

						else => error.InvalidSPIRV,
					},

					else => error.InvalidSPIRV,
				},
			};

			// uniform_block.uniforms[uniform_blocks_used].name = ext.spvc_compiler_get_member_name(compiler, ext.spvc_type_get_base_type_id(uniform_buffer_type), uniform_blocks_used);
		}
	}
}

fn to_error(result: ext.spvc_result) Error!void {
	return switch (result) {
		ext.SPVC_SUCCESS => {},
		ext.SPVC_ERROR_INVALID_SPIRV => error.InvalidSPIRV,
		ext.SPVC_ERROR_UNSUPPORTED_SPIRV => error.UnsupportedSPIRV,
		ext.SPVC_ERROR_OUT_OF_MEMORY => error.OutOfMemory,
		ext.SPVC_ERROR_INVALID_ARGUMENT, ext.SPVC_ERROR_INT_MAX => unreachable,
		else => unreachable,
	};
}
